import re
from abc import ABC
from abc import abstractmethod
from pathlib import Path


class _AbstractLogItem(ABC):
    """
    :class:`LogItem` interface, defining the attributes required by any :class:`LogItem` subclass.

    Methods
    -------
    parse(idx, file_path, item)
        parse a single log item, to produce the desired string representation in the output log.
    """

    @property
    @abstractmethod
    def search_patterns(self):
        """abstract property. A list of regex patterns used to search for the item."""
        pass

    @property
    @abstractmethod
    def template_marker(self):
        """
        abstract property. The marker used for insertion into a template. Built-in items use
        curly braces (e.g. ``{ assumptions }``).
        """
        pass

    @property
    @abstractmethod
    def empty_message(self):
        """
        abstract property. The text to be inserted into the log when no items are found. For
        example, 'No pokemon references found in this analysis.'.
        """
        pass

    @abstractmethod
    def parse(self, idx, file_path, item):
        """
        abstract method. Parse a log item into output string for writting to output log. Should
        return the output for a single log item, where each item is a match
        on the item's :attribute:`search_pattern`.

        Parameters
        ----------
        idx
            The index of the item in the list of captured items
        file_path
            The relative path to the file where the item is found
        item
            An item matched using :attribute:search_pattern`. A string, or tuple of strings,
            depending on number of groups captured in :attribute:search_pattern`.

        Returns
        -------
        str
            String representation of item, for use in output log file.
        """
        pass


class LogItem(_AbstractLogItem):
    """
    :class:`LogItem` interface, defining the attributes required by any :class:`LogItem` subclass.

    Attributes
    ----------
    matched_items
        list of log item matches that have been found.
    parsed_items
        list of parsed log items, which can be inserted into log outputs.

    Methods
    -------
    find_items(text, path)
        search for and store log items from text.
    parse_items()
        parse matched log items into strings.
    """

    def __init__(self):
        self.matched_items = []
        self.parsed_items = []

    def find_items(self, text: str, path: Path):
        """
        Search for log items in text. Stores matched items and their file
        paths in :attribute:`parsed_items`.

        Parameters
        ----------
        text
            a string of text to be searched for log items.
        path
            path to file containing text, for use in parsing.
        """
        for pattern in self.search_patterns:
            for item in re.findall(
                pattern,
                text,
                re.MULTILINE | re.IGNORECASE,
            ):
                self.matched_items.append((path, item))

    def parse_items(self):
        """
        parse each matched item into a string, ready to be inserted into output
        content.
        """
        self.parsed_items += [
            self.parse(idx, filepath, item)
            for idx, (filepath, item) in enumerate(self.matched_items)
        ]
        self.matched_items = []


class Assumption(LogItem):
    """
    Matches and parses assumptions from hash code comments.
    """

    search_patterns = [
        (
            # Genereric hash comments
            # Get indentation level and Assumption title
            r"^([ \t]*)# ?Assumption: ?(.+)\n"
            # Short or long from RAG ratings
            r"^\1# ?Q(?:uality)?: ?(.+)\n"
            r"^\1# ?I(?:mpact)?: ?(.+)\n"
            # Lazily match everything following, till there's a line that doesn't
            # start with the same indent and comment
            r"(\1# ?(?:.|\n)*?)^(?!\1#)"
        ),
        (
            # R roxygen format
            # As above, using #' notation
            r"^([ \t]*)#' @section Assumption: ?(.+)\n"
            r"^\1#' Q(?:uality)?: ?(.+)\n"
            r"^\1#' I(?:mpact)?: ?(.+)\n"
            r"(\1#' (?:.|\n)*?)^(?!\1#')"
        ),
    ]

    template_marker = "{ assumptions }"
    empty_message = "Currently no assumptions in this analysis.\n"

    def parse(self, idx, file_path, item):
        detailed_description = re.sub(
            # Remove indentation and comment hash from detailed description
            f"\n?{item[0]}#'? ?",
            "\n",
            item[4],
        )
        detailed_description = re.sub(
            # Reduce whitespace to single spaces and strip
            "[ ]{2,}",
            " ",
            detailed_description.strip(),
        )

        assumptions_content = "\n".join(
            [
                f"### Assumption {idx + 1}: {item[1]}",
                "",
                # Relative path to file
                f"* Location: `{file_path}`",
                f"* **Quality**: {item[2]}",
                f"* **Impact**: {item[3]}",
                "",
                f"{detailed_description}",
                "",
            ],
        )
        return assumptions_content


class Caveat(LogItem):
    """
    Matches and parses caveats from hash code comments.
    """

    search_patterns = [
        (
            # Generic hash comments
            # Get indentation level and Caveat title
            r"^([ \t]*)# ?Caveat: ?(.+)\n"
            # Lazily match everything following, till there's a line that doesn't start
            # with the same indent and comment
            # Long description is optional, as caveats might be one liners
            r"(\1# ?(?:.|\n)*?)?^(?!\1#)"
        ),
        (
            # R roxygen format
            r"^([ \t]*)#\' @section Caveat: ?(.+)\n"
            r"(\1#\' (?:.|\n)*?)?^(?!\1#\')"
        ),
    ]

    template_marker = "{ caveats }"
    empty_message = "Currently no caveats in this analysis.\n"

    def parse(self, idx, file_path, item):
        detailed_description = re.sub(
            # Remove indentation and comment hash from detailed description
            f"\n?{item[0]}#'? ?",
            "\n",
            item[2],
        )
        detailed_description = re.sub(
            # Reduce whitespace to single spaces and strip
            "[ ]{2,}",
            " ",
            detailed_description.strip(),
        )

        caveat_content = "\n".join(
            [
                f"### Caveat {idx + 1}: {item[1]}",
                "",
                # Relative path to file
                f"Location: `{file_path}`",
                "",
                f"{detailed_description}",
                "",
            ],
        )
        return caveat_content


class Debt(LogItem):
    """
    Matches and parses technical debt items from hash code comments.
    """

    search_patterns = [
        (
            # Get indentation level and title
            r"^([ \t]*)# ?Debt: (.+)\n?"
            # Lazily match everything following, till there's a line that doesn't start
            # with the same indent and comment
            r"(\1# ?(?:.|\n)*?)^(?!\1#)"
        ),
        (
            # R roxygen format
            r"^([ \t]*)#\' @section Debt: ?(.+)\n"
            r"(\1#\' (?:.|\n)*?)?^(?!\1#\')"
        ),
    ]

    template_marker = "{ debt }"
    empty_message = "Looks like we're debt free!\n"

    def parse(self, idx, file_path, item):
        debt_item = re.sub(
            # Remove indentation and comment hash from todo item
            f"\n?{item[0]}#'? ?",
            "\n",
            item[2],
        )
        debt_item = re.sub(
            # Reduce whitespace to single spaces and strip
            "[ ]{2,}",
            " ",
            debt_item.strip(),
        )

        debt_content = "\n".join(
            [
                f"### Debt {idx + 1}: {item[1]}",
                "",
                # Relative path to file
                f"Location: `{file_path}`",
                "",
                f"{debt_item}",
                "",
            ],
        )
        return debt_content


class Todo(LogItem):
    """
    Matches and parses todos from hash code comments.
    """

    search_patterns = [
        (
            # Get indentation level
            r"^([ \t]*)# ?TODO: (.+)\n?"
            # Lazily match everything following, till there's a line that doesn't start
            # with the same indent and comment
            # Long description is optional, as todos are often one liners
            r"(\1# ?(?:.|\n)*?)?^(?!\1#)"
        ),
    ]

    template_marker = "{ todos }"
    empty_message = "Great, there's nothing to do!\n"

    def parse(self, idx, file_path, item):
        todo_item = item[1] + re.sub(
            # Remove indentation and comment hash from todo item
            f"\n?{item[0]}#",
            "",
            item[2],
        )
        todo_item = re.sub(
            # Reduce whitespace to single spaces and strip
            "[ ]{2,}",
            " ",
            todo_item.strip(),
        )

        return f"- [ ] {todo_item}"



#!/usr/bin/env python
import argparse

from assumptions.log import _BUILTIN_ITEM_TYPES
from assumptions.log import Log


def cli():
    """Assumptions command line interface entry point."""
    # Parse args
    parser = argparse.ArgumentParser(
        description="Generate an assumptions and caveats log from code comments.",
        epilog="Your assumptions and caveats documented.",
    )
    parser.add_argument(
        "-l",
        "--log-type",
        type=str,
        default="assumptions_caveats_log",
        help="type of log to produce. Default is the" " assumptions and caveats log.",
    )
    parser.add_argument(
        "-o",
        "--outfile",
        type=str,
        default=None,
        help="output Markdown (`.md`) file path. Default is " "'LOG_TYPE.md'.",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        default="",
        help="relative path to directory to seach for"
        " assumptions and caveats under. Default is current"
        " directory.",
    )
    parser.add_argument(
        "-e",
        "--extension",
        type=str,
        default="",
        help="file extension to search for. Searches all file"
        " extensions by default.",
    )
    parser.add_argument(
        "-t",
        "--template",
        type=str,
        default=None,
        help="path to custom log template. Overrides template" " from log argument.",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="when flag is passed, no output log is generated.",
    )
    args = parser.parse_args()

    outfile = args.outfile or f"{args.log_type}.md"

    # Generate log
    log = Log(args.log_type, outfile)

    item_types = _BUILTIN_ITEM_TYPES.get(args.log_type)
    if item_types is None:
        raise ValueError(
            f"{args.log_type} is not a build in log type. Try one of: {','.join(_BUILTIN_ITEM_TYPES.keys())}",
        )

    for item_type_class in item_types:
        log.add_log_item_type(item_type_class)

    log.find_items(args.path, args.extension)
    updated = log.write_log(args.template, args.dry_run)
    if args.log_type == "assumptions_caveats_log":
        if not updated:
            print("\nNUDGE: Have you updated your assumptions and caveats?")
        else:
            print("\nAssumptions and caveats documented.")


if __name__ == "__main__":
    cli()



import datetime
import os
import re
from pathlib import Path

import pkg_resources

from assumptions.log_items import Assumption
from assumptions.log_items import Caveat
from assumptions.log_items import Debt
from assumptions.log_items import LogItem
from assumptions.log_items import Todo


class LogError(Exception):
    pass


class LogFindError(Exception):
    pass


_BUILTIN_ITEM_TYPES = {
    "assumptions_caveats_log": [Assumption, Caveat],
    "technical_debt_log": [Debt],
    "todo_list": [Todo],
}


class Log:
    """
    Searches files for log items and writes them to an output log using a
    template.

    Methods
    -------
    add_log_item_type(log_item)
        add a :class:`LogItem` subclass, for use when searching for log items.
    find_items(relative_search_path='', extension='')
        recursively search files under the specificed path for log items.
        Current directory and all file extensions by default.
    write_log(template_path=None, encoding: str='utf-8')
        write an output log file, inserting discovered log items into the
        provided template. Uses the built-in template for the choses log type
        by default.
    """

    def __init__(
        self,
        log_type: str = "assumptions_caveats_log",
        log_file_path: str = "assumptions_caveats_log.md",
    ):
        self._log_file_path = Path(log_file_path)
        if not self._log_file_path.parent.exists():
            raise FileNotFoundError(
                f"Output directory does not exist: {self._log_file_path.parent}",
            )

        self._log_item_types = []

        if log_type not in _BUILTIN_ITEM_TYPES.keys():
            msg = (
                f"{log_type} is not a valid log type."
                f" Choose from {', '.join(_BUILTIN_ITEM_TYPES.keys())}."
            )
            raise ValueError(msg)

        self._builtin_template = pkg_resources.resource_filename(
            "assumptions",
            f"templates/{log_type}.md",
        )

    def add_log_item_type(self, log_item: LogItem):
        """
        Add a :class:`LogItem` subclass to the log. These parsers provide the regex pattern for searching
        for items, the marker for insertion into templates and a handler method
        for parsing items before inserting them into the template.

        Parameters
        ----------
        log_item
            log items to search for and include in output log.
        """
        if not issubclass(log_item, LogItem):
            raise TypeError(
                "Log item must be a subclass of `assumptions.LogItem`",
            )
        self._log_item_types.append(log_item())

    def find_items(self, relative_search_path: str = "", extension: str = ""):
        """
        Recursive directory search for the :attribute:`search_pattern` of each :attribute:`log_item`.
        Optionally searches a specific file extension. Also provides the
        relative path to current file.

        Parameters
        ----------
        relative_search_path
            relative path to directory that should be searched for log items.
        extension
            file extension to reduce search to specific file types (e.g. '.py').
        """
        if len(self._log_item_types) == 0:
            raise LogError("No `log_items` have been added to the Log.")

        current_dir = Path(os.getcwd())
        search_path = (current_dir / relative_search_path).resolve()
        print(f"Searching for log items under: {search_path}")

        for path in [p for p in search_path.glob("**/*" + extension) if p.is_file()]:
            try:
                with path.open("r") as f:
                    file_contents = f.read()
            except UnicodeDecodeError:
                print(f"File could not be read, skipping: {path}")

            for log_item in self._log_item_types:
                log_item.find_items(
                    file_contents,
                    path.relative_to(search_path.parent).as_posix(),
                )

    def write_log(
        self,
        template_path: str = None,
        dry_run: bool = False,
        encoding: str = "utf-8",
    ):
        """
        Write log to instance :attribute:`log_file_path`. Inserts matched log items into
        markers in the specified template file.

        Paramters
        ---------
        template_path
            path to the output log template (any plain text file).
        encoding
            encoding used to read and write template and output log.
        """
        if template_path is None:
            # Default is assumptions and caveats from package
            template_path = self._builtin_template
        with open(template_path, "r", encoding=encoding) as f:
            template_content = f.read()

        if "{ current_date }" in template_content:
            template_content = template_content.replace(
                "{ current_date }",
                datetime.datetime.today().strftime(r"%d/%m/%Y"),
            )

        for log_item_type in self._log_item_types:
            log_item_type.parse_items()
            items = log_item_type.parsed_items

            if len(items) == 0:
                print(f"Warning: No {log_item_type.__class__.__name__} items found.")
                items = [log_item_type.empty_message]

            template_content = template_content.replace(
                log_item_type.template_marker,
                "\n".join(items).strip(),
            )

        if self._log_file_path.exists():
            print("Log exists, checking for changes...")
            with open(self._log_file_path, "r", encoding=encoding) as f:
                old_template_content = f.read()

            # Check if output has changed, other than dates
            date_format = r"[0-9]{2}/[0-9]{2}/[0-9]{4}"
            if re.sub(date_format, "", old_template_content) == re.sub(
                date_format,
                "",
                template_content,
            ):
                print("No change to log items, log not updated.")
                return False

        if not dry_run:
            print(f"Writing log to: {self._log_file_path}")
            with open(self._log_file_path, "w", encoding=encoding) as f:
                f.write(template_content)
        return True



from pathlib import Path

from setuptools import setup

# Get version from package __init__
text = Path("./assumptions/__init__.py").read_text(encoding="utf8")
for line in text.split("\n"):
    if "__version__" in line:
        break
version = line.split("= ")[-1].strip('"')

path_doc_reqs = Path(__file__).parent.joinpath("docs", "requirements.txt")
testing_reqs = [
    "coverage",
    "pytest>=3.6,<4",
    "pytest-cov",
    "pytest-regressions",
]


def read_requirements(file):
    with open(file) as f:
        return f.read().splitlines()


setup(
    name="assumptions",
    version=version,
    description="Generating Markdown assumptions logs from code comments.",
    long_description=Path("./README.rst").read_text(encoding="utf8"),
    long_description_content_type="text/x-rst",
    author="David Foster",
    author_email="foster.dev999@gmail.com",
    packages=["assumptions"],
    include_package_data=True,
    license="MIT",
    python_requires=">=3.6",
    extras_require={
        "testing": testing_reqs,
        "docs": read_requirements(path_doc_reqs),
        "dev": ["bump2version"] + testing_reqs,
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python",
        "Topic :: Documentation",
        "Topic :: Software Development :: Documentation",
        "Topic :: Text Processing",
    ],
    entry_points={
        "console_scripts": ["assumptions=assumptions.cli:cli"],
    },
)
